@@ -35,8 +35,13 @@ typedef struct {
 	char           *description;
 	AccessibleRole  role;
 	char           *role_name; /* sad but true */
+	gboolean        display_children;
 } ModelNode;
 
+int max_children = 0;
+gboolean dontdisplaychildren = FALSE;
+gboolean use_table_if = FALSE;
+
 static GNode *
 model_node_new (AccessibleTreeModel *model, Accessible *accessible)
 {
@@ -47,6 +52,17 @@ model_node_new (AccessibleTreeModel *model, Accessible *accessible)
 	mnode->accessible = accessible;
 	Accessible_ref (accessible);
 
+	mnode->display_children = TRUE;
+	if (dontdisplaychildren) {
+		AccessibleStateSet *state_set;
+
+		state_set = Accessible_getStateSet (accessible);
+		if (AccessibleStateSet_contains (state_set,
+						 SPI_STATE_MANAGES_DESCENDANTS)) {
+			mnode->display_children = FALSE;
+		}
+		AccessibleStateSet_unref (state_set);
+	}
 	node = g_node_new (mnode);
 
 	g_hash_table_insert (model->accessible_to_node, accessible, node);
@@ -153,14 +169,42 @@ model_node_destroy (GtkTreeModel *model, GNode *node)
 	gtk_tree_path_free (path);
 }
 
+static gboolean
+model_node_has_children (GtkTreeModel *tree_model, GNode *node)
+{
+	ModelNode *mnode;
+	AccessibleTreeModel *model = ACCESSIBLE_TREE_MODEL (tree_model);
+	Accessible *accessible;
+	gboolean retval;
+
+	g_return_val_if_fail (node != NULL, FALSE);
+
+	mnode = node->data;
+
+	if (!mnode->display_children)
+		return FALSE;
+
+	if (mnode->state & NODE_GOT_CHILDREN)
+		return node->children != NULL;
+
+	accessible = Accessible_getChildAtIndex (mnode->accessible, 0);
+	retval = accessible != NULL;
+	if (accessible)
+		Accessible_unref (accessible);
+
+	return retval;
+}
+
 static void
 model_node_sync_children (GtkTreeModel *tree_model, GNode *node)
 {
-	int i;
+	int i, j;
 	GNode  *lnode;
 	GSList *children, *l;
 	ModelNode *mnode;
 	AccessibleTreeModel *model = ACCESSIBLE_TREE_MODEL (tree_model);
+	AccessibleTable *accessible_table;
+	gboolean use_table = FALSE;
 
 	g_return_if_fail (node != NULL);
 
@@ -169,28 +213,60 @@ model_node_sync_children (GtkTreeModel *tree_model, GNode *node)
 	if (mnode->state & NODE_GOT_CHILDREN)
 		return;
 
+	if (use_table_if) {
+		accessible_table = Accessible_getTable (mnode->accessible);
+
+		if (accessible_table)
+			use_table = TRUE;
+	} 			
+
  retry_on_reenter:
 	mnode->state |= NODE_GOT_CHILDREN | NODE_FETCHING_CHILDREN;
 
 	model->stamp++;
 	children = NULL;
-	for (i = 0; ; i++) {
-		Accessible *accessible;
 
-		accessible = Accessible_getChildAtIndex (mnode->accessible, i);
+	if (use_table) {
+		for (i = 0; max_children ? i < max_children : TRUE; i++) {
+			for (j = 0; max_children ? j < max_children : TRUE; j++) {
+				Accessible *accessible;
 
-		if (mnode->state & NODE_CHILDREN_CHANGED) {
-			mnode->state &= ~NODE_CHILDREN_CHANGED;
-			for (l = children; l; l = l->next)
-				Accessible_unref (l->data);
-			g_slist_free (children);
-			goto retry_on_reenter;
+				accessible = AccessibleTable_getAccessibleAt (accessible_table, i, j);
+				if (mnode->state & NODE_CHILDREN_CHANGED) {
+					mnode->state &= ~NODE_CHILDREN_CHANGED;
+					for (l = children; l; l = l->next)
+						Accessible_unref (l->data);
+					g_slist_free (children);
+					goto retry_on_reenter;
+				}
+
+				if (!accessible)
+					break;
+
+				children = g_slist_prepend (children, accessible);
+			}
 		}
+		Accessible_unref (accessible_table);
+	} else {
+	
+		for (i = 0; max_children ? i < max_children : TRUE; i++) {
+			Accessible *accessible;
 
-		if (!accessible)
-			break;
+			accessible = Accessible_getChildAtIndex (mnode->accessible, i);
 
-		children = g_slist_prepend (children, accessible);
+			if (mnode->state & NODE_CHILDREN_CHANGED) {
+				mnode->state &= ~NODE_CHILDREN_CHANGED;
+				for (l = children; l; l = l->next)
+					Accessible_unref (l->data);
+				g_slist_free (children);
+				goto retry_on_reenter;
+			}
+
+			if (!accessible)
+				break;
+
+			children = g_slist_prepend (children, accessible);
+		}
 	}
 	mnode->state &= ~NODE_FETCHING_CHILDREN;
 
@@ -226,6 +302,7 @@ model_node_sync_children (GtkTreeModel *tree_model, GNode *node)
 			GNode *child;
 
 			child = model_node_new (model, l->data);
+			Accessible_unref (l->data);
 
 			g_node_insert_before (node, lnode, child);
 			model_node_inserted (tree_model, child);
@@ -236,12 +313,23 @@ model_node_sync_children (GtkTreeModel *tree_model, GNode *node)
 }
 
 static void
-model_node_children_changed (AccessibleTreeModel *tree_model, GNode *node)
+model_node_children_changed (AccessibleTreeModel *tree_model, Accessible *source)
 {
+	GNode *node;
 	ModelNode *mnode;
 
+	node = g_hash_table_lookup (tree_model->accessible_to_node, source);
+
 	if (!node) {
-		g_warning ("Changed in another tree model !");
+		/*
+		 * This is not neccesarily an error - quite possibly
+		 * we're getting events from a tree we're not interested
+		 * in or a tree which is in the process of being deleted
+		 * which is quite ok.
+		 */
+#if DEBUG_TREE
+		g_warning ("Changed ('%p') in another tree model !", source);
+#endif
 		return;
 	}
 
@@ -250,6 +338,9 @@ model_node_children_changed (AccessibleTreeModel *tree_model, GNode *node)
 		return;
 	}
 
+	if (!mnode->display_children)
+		return;
+
 	if (mnode->state & NODE_FETCHING_CHILDREN) {
 		mnode->state |= NODE_CHILDREN_CHANGED;
 		g_warning ("Heavy duty re-entrancy on node '%p'", node);
@@ -451,8 +542,7 @@ accessible_tree_model_iter_has_child (GtkTreeModel *model,
 	GNode *node = ITER_NODE (iter);
 
 	if (node) {
-		model_node_sync_children (model, node);
-		return node->children != NULL;
+		return model_node_has_children (model, node);
 	} else
 		return FALSE;
 }
@@ -511,13 +601,19 @@ accessible_tree_model_unref_node (GtkTreeModel *tree_model, GtkTreeIter *iter)
 {
 	GNode *node = ITER_NODE (iter);
 	ModelNode *mnode;
+	AccessibleTreeModel *model;
 
 	if (!node || !(mnode = node->data))
 		return;
 
 	mnode->ref--;
-	if (mnode->ref == 0)
-		/* FIXME: leak here ? */ ;
+	if (mnode->ref == 0) {
+		model = ACCESSIBLE_TREE_MODEL (tree_model);
+		if (model->root == node) {
+			model_node_destroy (tree_model, node);
+			model->root = NULL;
+		}
+	}
 }
 
 gboolean
@@ -532,8 +628,27 @@ accessible_is_base_accessible (Accessible *accessible)
 
 	CORBA_exception_init (&ev);
 	ret = CORBA_Object_is_a (*a, "IDL:Accessibility/Accessible:1.0", &ev);
-	if (BONOBO_EX (&ev))
+	if (BONOBO_EX (&ev)) {
+#ifdef JAVA_ENABLED
+		/*
+		* The Java ORB fails when an "is_a" invocation is sent over the
+		* wire with an exception BAD_OPERATION. 
+		* Because of this, we need to make an attempt at using something
+		* else to determin if the base is Accessible or not.
+		*/
+		Accessible *aa;
+		
+		aa = Accessible_queryInterface (accessible, "IDL:Accessibility/Accessible:1.0");
+		if ( aa != NULL ) {
+		    ret = TRUE;
+		    Accessible_unref (aa);
+		}
+		else
+		    ret = FALSE;
+#else /* JAVA_ENABLED */
 		ret = FALSE;
+#endif /* JAVA_ENABLED */
+	}
 	CORBA_exception_free (&ev);
 
 	return ret;
@@ -547,9 +662,8 @@ accessible_tree_model_new (Accessible *root)
 	g_return_val_if_fail (root != NULL, NULL);
 	
 	model = g_object_new (ACCESSIBLE_TYPE_TREE_MODEL, NULL);
-#ifndef JAVA_ENABLED
+
 	g_assert (accessible_is_base_accessible (root)); 
-#endif
 
 	model_construct (ACCESSIBLE_TREE_MODEL (model), root);
 
@@ -561,14 +675,25 @@ accessible_tree_model_new (Accessible *root)
 static void
 report_global_event (const AccessibleEvent *event, void *user_data)
 {
-	GNode *node;
 	AccessibleTreeModel *model = ACCESSIBLE_TREE_MODEL (user_data);
 
 	if (!model || !model->accessible_to_node)
 		return;
 
-	node = g_hash_table_lookup (model->accessible_to_node, event->source);
-	model_node_children_changed (model, node);
+	if (!strncmp (event->type, "object:children-changed", 
+                           strlen( "object:children-changed") )) {
+		model_node_children_changed (model, event->source);
+	} else if (!strcmp (event->type, "object:active-descendant-changed")) {
+		Accessible *active_descendant;
+		char *role_txt;
+
+		active_descendant = AccessibleActiveDescendantChangedEvent_getActiveDescendant (event);
+
+		role_txt = Accessible_getRoleName (active_descendant);
+
+		g_print ("New active descendant: %p Role: %s Index: %d\n", active_descendant, role_txt, event->detail1);
+		SPI_freeString (role_txt);
+	}
 }
 
 static void
@@ -580,6 +705,8 @@ accessible_tree_model_init (AccessibleTreeModel *model)
 		report_global_event, model);
 	SPI_registerGlobalEventListener (
 		model->event_listener, "object:children-changed");
+	SPI_registerGlobalEventListener (
+		model->event_listener, "object:active-descendant-changed");
 }
 
 static void
@@ -588,6 +715,8 @@ accessible_tree_model_dispose (GObject *object)
 	AccessibleTreeModel *model;
 
 	model = ACCESSIBLE_TREE_MODEL (object);
+	if (model->root)
+		model_node_destroy (GTK_TREE_MODEL (model), model->root);
 
 	if (model->event_listener != NULL) {
 		SPI_deregisterGlobalEventListenerAll (
